// $Id: StdServiceClient.java,v 1.6 2006/08/31 08:00:57 orts_mburo Exp $

// (c) Michael Buro, licensed under the GPL2

package ggs.shared;

import java.util.*;

public abstract class StdServiceClient extends ServiceClient
{
  /** initialize */
  public void init(Sender sender, Options opt)
  {
    super.init(sender, opt);
  }

  /** service arrived */
  public void arrived()
  {
    super.arrived();

    send("t " + getLogin() + " list"); // trigger sending table list
    
    // send initial commands to service
    String cmds = options.getString(getOptPrefix() + ".cm");

    if (cmds == null) return;

    if (cmds.length() > 0) {
      StringTokenizer st = new StringTokenizer(cmds, ";");
      while (st.hasMoreTokens()) {
	send("t " + getLogin() + " " + st.nextToken());
      }
    }
  }

  /** service departed */
  public void departed()
  {
    super.departed();

    // erase all data
    joinedTables.clear();
    tableInfos.clear();
  }

  /** clean up */
  public void shut_down()
  {
  }

  /** append service options to Global.options */
  public void setDefaultOptions()
  {
  }

  /** option prefix */
  protected abstract String getOptPrefix();

  protected abstract Game getGameBP();

  public boolean handle_event(EventMsg msg)
  {
    if (msg.get_args().size() == 0)
      return false;

    // handle only messages for this service
    if (!msg.get_args().get(0).startsWith(getLogin()))
      return false;

    for (String currArg : msg.get_args()) {
      // print out all params
      Misc.msg("::: " + currArg);
    }

    SReader first = new SReader(msg.get_args().get(0)+""); // clone
      
    String login   = first.nextWord();
    String command = first.nextWord();
    
    // copy remaining lines and remove first
    Vector<String> rest = new Vector<String>(msg.get_args());
    rest.remove(0);
    
    for (StdServiceMsgData e : exec) {
      if (command.equals(e.getMsg())) {
	e.getExec().exec(first, rest);
	return true;
      }
    }

    return false;
  }

  /*
   * inner classes for handling table messages
   */

  /**
   * Message interface
   */
  abstract class StdServiceMsg
  {
    /**
     * handle service message
     * @param first tokenized first line
     * @lines all input lines
     */
    abstract void exec(SReader first, Vector<String> lines);
  }

  /**
   * Element of message list (see below)
   */
  class StdServiceMsgData
  {

    public StdServiceMsgData(String msg, StdServiceMsg exec)
    {
      this.msg = msg;
      this.exec = exec;
    }

    public String getMsg() { return msg; }
    public StdServiceMsg getExec() { return exec; }
    private String msg;
    private StdServiceMsg exec;
  }

  // the list of recognized service message prefixes
  private StdServiceMsgData exec[] = {
    new StdServiceMsgData("list",  new MsgList()),  // list of table info changes
    new StdServiceMsgData("state", new MsgState()), // whole table state
    new StdServiceMsgData("table", new MsgTable())  // table message prefix
  };

  /*
   * state <table-id> <table-state>
   * 
   */
  class MsgState extends StdServiceMsg
  {
    public void exec(SReader first, Vector<String> rest)
    {
      // read table state from string

      StdTable table = (StdTable)getGameBP().newTable();
      String tid = first.nextWord();

      if (tid == null) {
	Misc.err("no <tid> in state");
	return;
      }

      table.fromSReader(first);

      // add or replace table
      joinedTables.put(tid, table);
    }
  }

  
  /**
   * table messages
   *
   * table <id> <cmd> ...
   */
    
  public class MsgTable extends StdServiceMsg
  {
    public void exec(SReader first, Vector<String> rest)
    {
      Misc.msg("StdServiceClient table");

      // find table and delegate messages to it

      Integer id = Integer.parseInt(first.nextWord());
      StdTable tab = joinedTables.get(id);
      
      if (tab == null) {
	Misc.msg("unknown table - should not happen");
	return;
      }

      tab.execInClient(first, rest);
    }
  }


  /**
   *  Tableinfo update
   *
   *  list
   *  + <tid> <table-info>
   *  - <tid>
   *  ...
   */
  public class MsgList extends StdServiceMsg
  {

    public void exec(SReader first, Vector<String> rest)
    {
      Misc.msg("StdServiceClient list");

      for (String s : rest) {

	SReader  r = new SReader(s);
	String  op = r.nextWord();
	String tid = r.nextWord();
	
	if (op.equals("+")) {

	  // add table info

	  StdTableInfo ti = new StdTableInfo();
	  ti.fromString(r.rest());
	  tableInfos.put(tid, ti);

	} else if (op.equals("-")) {

	  tableInfos.remove(tid);
	  
	} else {

	  Misc.msg("WARNING: list operation " + op + "?");
	  return;
	}
      }
    }
  }

  /** joined tables (maps ids to infos) */
  public HashMap<String, StdTable> joinedTables = new HashMap<String, StdTable>();

  /** infos for all tables (maps ids to infos) */
  public HashMap<String, StdTableInfo> tableInfos = new HashMap<String, StdTableInfo>();

  // game blueprint for generating tables  
  Game gameBP; 
}
